unit LKCheckGroupBox;

interface

uses
	Classes,
	Controls,
	Messages,
	StdCtrls;

type
	TLKCheckGroupBox = class;

	TLKCheckBoxList = class(TList)
		Owner: TLKCheckGroupBox;
		constructor Create(AOwner: TLKCheckGroupBox);
		function ReadItem(Index: Integer): TCheckBox;
		procedure WriteItem(Index: Integer; Value: TCheckBox);
		function Add(Item: TCheckBox): TCheckBox;
		function LastIndex: Integer;
		procedure Delete(Index: Integer);
		procedure Remove(Item: TCheckBox);
		procedure Clear; override;
		property Items[Index: Integer]: TCheckBox read ReadItem write WriteItem; default;
		destructor DestroyOwned;
	end;

	TNotifyStateChange = procedure(Sender: TObject; Box: TCheckBox) of object;
	NOrdering = (orderVert, orderHoriz);
	TLKCheckGroupBox = class(TGroupBox)
	private
		fXList: TLKCheckBoxList;
		fXNameList: TStrings;
		fColumns: Integer;
		fRowColRatio: Double;
		fDefState: TCheckBoxState;
		fGrayed: Boolean;
		fBoxOrder: NOrdering;
		fReading: Boolean;
		fProcessingClick: Boolean;
		fOnChecked: TNotifyStateChange;
		fOnUnchecked: TNotifyStateChange;
		fOnChanged: TNotifyStateChange;
		function GetNameList: TStrings;
		procedure SetNameList(Value: TStrings);
		procedure SetRowColRatio(Value: Double);
		function GetRowColRatio: Double;
		procedure SetColumns(Value: Integer);
		procedure SetGrayed(Value: Boolean);
		procedure XListUpdated;
		procedure SetBoxOrder(Value: NOrdering);
		procedure SetDefState(Value: TCheckBoxState);
		procedure CheckBoxClick(Sender: TObject);
		procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
		procedure WMSize(var Message: TWMSize); message WM_SIZE;
	protected
		procedure ReadState(Reader: TReader); override;
		procedure Loaded; override;
		procedure SetPositions;
	public
		constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;
		//Add a single Check-Box to the Group
		function AddItem(Name: string): Integer;
		//Set the state of all/a subset of boxes in the group
		procedure SetBoxes(State: TCheckBoxState; BoxList: TList);
		//Invert the state of all/a subset of boxes
		procedure InvertBoxes(BoxList: TList; IncludeGrayed: Boolean);
		//return a subset of the boxes in the group
		procedure GetBoxes(State: TCheckBoxState; var BoxList: TLKCheckBoxList); overload;
		//return a subset of box-names/-captions
		procedure GetNames(State: TCheckBoxState; var Names: TStrings); overload;
		//return all boxes in the group
		procedure GetBoxes(var BoxList: TLKCheckBoxList); overload;
		//return all box-names/captions in the group
		procedure GetNames(var Names: TStrings); overload;
		//provides run-time access to the existing boxes
		property Items: TLKCheckBoxList read fXList stored FALSE;
		property Canvas;
	published
		//get/set the names of all boxes in the group
		property ItemNames: TStrings read GetNameList write SetNameList;
		//number of columns
		property Columns: Integer read fColumns write SetColumns;
		//CheckState of all newly or recreated boxes
		property DefaultState: TCheckBoxState read fDefState write SetDefState;
		//similar to TCheckBox.AllowGrayed
		property AllowGrayed: Boolean read fGrayed write SetGrayed;
		//line the boxes up vertically or horizontally
		property BoxOrder: NOrdering read fBoxOrder write SetBoxOrder;
		//keep a fixed Ratio of RowCount to ColumnCount
		property RowToColRatio: Double read GetRowColRatio write SetRowColRatio;
		//notify, when boxes get Check-Mark set.
		property OnBoxesChecked: TNotifyStateChange read fOnChecked write fOnChecked;
		//notify, when boxes get Check-Mark reset.
		property OnBoxesUnchecked: TNotifyStateChange read fOnUnchecked write fOnUnchecked;
		//notify, when boxes state is changed.
		property OnBoxesChanged: TNotifyStateChange read fOnChanged write fOnChanged;
	end;



implementation

const
	boxXOffs = 10;
	boxYOffs = 5;
	boxWidth = 20;

{	*************************************************************************************
	*  class TLKCheckBoxList - Descendant of TList to contain TCheckboxes               *
	*************************************************************************************}

//if the class is created with a CheckGroupBox as owner, that Box is notified about
//	changes in the List.
constructor TLKCheckBoxList.Create(AOwner: TLKCheckGroupBox);
begin
	inherited Create;
	Owner := AOwner
end;

function TLKCheckBoxList.ReadItem(Index: Integer): TCheckBox;
begin
	Result := inherited Items[Index];
end;

procedure TLKCheckBoxList.WriteItem(Index: Integer; Value: TCheckBox);
begin
	inherited Items[Index] := Value;
	if Assigned(Owner) then
		Owner.XListUpdated;
end;

function TLKCheckBoxList.Add(Item: TCheckBox): TCheckBox;
var
	I: Integer;
begin
	I := inherited Add(Item);
	if Assigned(Owner) then
		Owner.XListUpdated;
	Result := Items[I];
end;

function TLKCheckBoxList.LastIndex: Integer;
begin
	Result := Count - 1;
end;

destructor TLKCheckBoxList.DestroyOwned;
begin
	Clear;
	if Assigned(Owner) then begin
		Owner.fXList := nil;
		Owner := nil;
	end;
	inherited Destroy;
end;

procedure TLKCheckBoxList.Delete(Index: Integer);
begin
	Items[Index].Free;
	inherited Delete(Index);
	if Assigned(Owner) then
		Owner.XListUpdated;
end;

procedure TLKCheckBoxList.Remove(Item: TCheckBox);
begin
	inherited Remove(Item);
	Item.Free;
	if Assigned(Owner) then
		Owner.XListUpdated;
end;

procedure TLKCheckBoxList.Clear;
begin
	if Assigned(Owner) then
	while Count > 0 do begin
		Items[LastIndex].Free;
		inherited Delete(LastIndex);
	end;
	inherited;
	if Assigned(Owner) then
		Owner.XListUpdated;
end;


{	*************************************************************************************
	*********   Klasse TLKCheckGroupBox - eine GroupBox, die Checkboxen enthlt    ********
	*************************************************************************************}

//*************************************************************
constructor TLKCheckGroupBox.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
	fXList := TLKCheckBoxList.Create(Self);
	fXNameList := TStringList.Create;
	fColumns := 1;
	fDefState := cbUnchecked;
	fGrayed := False;
	fBoxOrder := orderVert;
	fProcessingClick := False;
end;

//*************************************************************
destructor TLKCheckGroupBox.Destroy;
begin
	fXNameList.Free;
	fXList.DestroyOwned;
	inherited Destroy;
end;

//*************************************************************
//Handles notification by a change in the attached TCheckBoxList
procedure TLKCheckGroupBox.XListUpdated;
begin
	SetPositions;
end;

//*************************************************************
//public: Add a Checkbox with the passed Name.
function TLKCheckGroupBox.AddItem(Name: string): Integer;
var
	x: tCheckBox;
begin
	if fRowColRatio > 0 then
		fColumns := Round(SQRT((fXList.Count+1) / fRowColRatio));
	X := fXList.Add(TCheckBox.Create(Self));
	X.Parent := self;
	X.AllowGrayed := fGrayed;
	X.State := fDefState;
	X.Caption := Name;
	X.Width := Self.Canvas.TextWidth(Name) + boxWidth;
	X.Height := Self.Canvas.TextHeight(Name);
	X.Enabled := Self.Enabled;
	X.OnClick := Self.CheckBoxClick;
	Result :=	fXList.LastIndex;
end;

//*************************************************************
//Set no. of columns  (Columns.write)
procedure TLKCheckGroupBox.SetColumns(Value: Integer);
begin
	if (Value > 0) then begin
		fColumns := Value;
		fRowColRatio := -1;
	end;
	SetPositions;
end;

//*************************************************************
procedure TLKCheckGroupBox.SetRowColRatio(Value: Double);
begin
	if Value>0 then begin
		fRowColRatio := Value;
		fColumns := Round(SQRT(fXList.Count / fRowColRatio));
	end;
	SetPositions;
end;

//*************************************************************
function TLKCheckGroupBox.GetRowColRatio: Double;
begin
	if fRowColRatio = 0 then
		Result := fXList.Count / (Columns*Columns)
	else
		Result := fRowColRatio;
end;


//*************************************************************
// calculate and set the positions for each contained Checkbox
procedure TLKCheckGroupBox.SetPositions;
var
	I, NoRows, ColWid, RowHgt: Integer;
begin
	if (fXList.Count=0) or fReading then
		Exit;
	NoRows := fXList.Count div fColumns;
	if (fXList.Count mod fColumns)>0 then
		inc(NoRows);
	ColWid := self.ClientWidth div fColumns;
	RowHgt := (self.ClientHeight-fXList[0].Height) div NoRows;
	if fBoxOrder = orderVert then
	begin
		for I := 0 to fXList.LastIndex do
		with fXList[I] do begin
			Left := (I div NoRows)*ColWid + boxXOffs;
			Top := Trunc((I mod NoRows + 0.5)*RowHgt + boxYOffs);
		end;
	end
	else begin
		for I := 0 to fXList.LastIndex do
		with fXList[I] do begin
			Left := (I mod fColumns)*ColWid + boxXOffs;
			Top := Trunc((I div fColumns + 0.5)*RowHgt + boxYOffs);
		end;
	end;
end;

//*************************************************************
// Get the names (captions) of all contained CheckBoxes and
// return them as STringList (property ItemNames.read)
function TLKCheckGroupBox.GetNameList: TStrings;
var
	I: Integer;
begin
	if not Assigned(fXNameList) then
		fXNameList := TStringList.Create;
	fXNameList.Clear;
	for I := 0 to fXList.LastIndex do
		fXNameList.Add(fXList[I].Caption);
	Result := fXNameList;
end;

//*************************************************************
// Recreate all Chechkboxes in the Group using the Captions that are
// passed in the StringList (property ItemNames.write)
procedure TLKCheckGroupBox.SetNameList(Value: TStrings);
var
	I: Integer;
begin
	fXList.Clear;
	for I := 0 to Value.Count-1 do
		AddItem(Value[I]);
end;

//*************************************************************
//Returns a List of all Checkboxes in the Group (for manipulating single Checkboxes)
procedure TLKCheckGroupBox.GetBoxes(var BoxList: TLKCheckBoxList);
var
	I: Integer;
begin
	if not Assigned(BoxList) then
		BoxList := TLKCheckBoxList.Create(nil);
	with BoxList do begin
		Clear;
		for I := 0 to fXList.LastIndex do
			Add(fXList[I]);
	end;
end;

//*************************************************************
//Returns a List of those Checkboxes in the Group, that are set to
// the passed CheckBoxState
procedure TLKCheckGroupBox.GetBoxes(State: TCheckBoxState; var BoxList: TLKCheckBoxList);
var
	I: Integer;
begin
	if not Assigned(BoxList) then
		BoxList := TLKCheckBoxList.Create(nil);
	with BoxList do begin
		Clear;
		for I := 0 to fXList.LastIndex do
			if fXList[I].State = State then
				Add(fXList[I]);
	end;
end;

//*************************************************************
//Returns the Names (Captions) of all Checkboxes in the List
// (similar to property ItemNames.read, but as procedure with Parameter)
procedure TLKCheckGroupBox.GetNames(var Names: TStrings);
var
	I: Integer;
begin
	if not Assigned(Names) then
		Names := TStringList.Create;
	with Names do begin
		Clear;
		for I := 0 to fXList.LastIndex do
			Add(fXList[I].Caption);
	end;
end;

//*************************************************************
//Returns the Names (Captions) of all Checkboxes, that are set to
// the CheckBoxState passed in State.
procedure TLKCheckGroupBox.GetNames(State: TCheckBoxState; var Names: TStrings);
var
	I: Integer;
begin
	if not Assigned(Names) then
		Names := TStringList.Create;
	with Names do begin
		Clear;
		for I := 0 to fXList.LastIndex do
			if fXList[I].State = State then
				Add(fXList[I].Caption);
	end;
end;

//*************************************************************
// Sets the "AllowGrayed" property of all Checkboxes (Grayed.write)
procedure TLKCheckGroupBox.SetGrayed(Value: Boolean);
var
	I: Integer;
begin
	fGrayed := Value;
	for I := 0 to fXList.LastIndex do
		fXList[I].AllowGrayed := Value;
end;


//*************************************************************
//Rearranges the Boxes in the new Orientation (if changed).
// (BoxOrder.write).
procedure TLKCheckGroupBox.SetBoxOrder(Value: NOrdering);
begin
	if Value <> fBoxOrder then
	begin
		fBoxOrder := Value;
		SetPositions;
	end;
end;

//*************************************************************
// OnClick-Handler for all owned Checkboxes.
procedure TLKCheckGroupBox.CheckBoxClick(Sender: TObject);
var
	CB: TCheckBox;
begin
	if fProcessingClick then
		Exit;
	if Sender is TCheckBox then
		CB := TCheckBox(Sender)
	else
		Exit;
	if Assigned(fOnChanged) then
		fOnChanged(Self, CB);
	if (CB.State = cbChecked) and Assigned(fOnChecked) then
		fOnChecked(Self, CB)
	else if (CB.State = cbUnchecked) and Assigned(fOnUnchecked) then
		fOnUnchecked(Self, CB);
end;

//*************************************************************
//Rearranges the Boxes on Resizing of the Group
procedure TLKCheckGroupBox.WMSize(var Message: TWMSize);
begin
	inherited;
	SetPositions;
end;

//*************************************************************
// Sets the CheckState of all Boxes that are passed in BoxList
// to the passed State (BoxList can be a subset of all Boxes
// in the Group).
procedure TLKCheckGroupBox.SetBoxes(State: TCheckBoxState; BoxList: TList);
var
	I, N: Integer;
begin
	if Assigned(BoxList) then begin
		for I := 0 to BoxList.Count-1 do begin
			N := fXList.IndexOf(BoxList[I]);
			if N>=0 then
				fXList[N].State := State;
		end;
	end
	else
		for I := 0 to fXList.LastIndex do
			fXList[I].State := State;
end;

//*************************************************************
// Inverts the CheckState of all Boxes (cbChecked <-> cbUnchecked)
// that are passed in BoxList (BoxList can be a subset of all Boxes
// in the Group). If IncludeGrayed = TRUE, grayed boxes are treated like
// unchecked boxes.
procedure TLKCheckGroupBox.InvertBoxes(BoxList: TList; IncludeGrayed: Boolean);
var
	I, N: Integer;
begin
	if Assigned(BoxList) then begin
		for I := 0 to BoxList.Count-1 do begin
			N := fXList.IndexOf(BoxList[I]);
			if (N>=0) and (IncludeGrayed or (fXList[N].State <> cbGrayed)) then
				fXList[N].Checked := not fXList[N].Checked;
		end;
	end
	else
		for I := 0 to fXList.LastIndex do
			if IncludeGrayed or (fXList[I].State <> cbGrayed) then
				fXList[I].Checked := not fXList[I].Checked;

end;

//*************************************************************
//Sets the default state for newly created Boxes (DefaultState.write).
// Boxes are (re-)created via the Method "AddItem" or by setting
// the property "ItemNames" (see SetNameList = ItemNames.write).
procedure TLKCheckGroupBox.SetDefState(Value: TCheckBoxState);
var
	I: Integer;
begin
	fDefState := Value;
	if csDesigning in ComponentState then
		for I := 0 to fXList.LastIndex do
			fXList[I].State := Value;
end;

//*************************************************************
//Passes a change in the group's Enabled state to all contained Boxes.
procedure TLKCheckGroupBox.CMEnabledChanged(var Message: TMessage);
var
	I: Integer;
begin
	for I := 0 to fXList.LastIndex do
		fXList[I].Enabled := Enabled;
end;

//*************************************************************
//Disables multiple calls of "SetPositions" during reading of the
// component by a TReader
procedure TLKCheckGroupBox.ReadState(Reader: TReader);
begin
	fReading := True;
	inherited ReadState(Reader);
	fReading := False;
	SetNameList(fXNameList);
end;

//*************************************************************
//Sets the Positions of the Boxes after the Component has been read.
procedure TLKCheckGroupBox.Loaded;
begin
	inherited Loaded;
	SetPositions;
end;

end.
